import base64
import re
import zipfile

import lib.generic
from lxml import etree


class QTIConverter(lib.generic.Converter):
    """ XLSConverter class (Currently raw xml input is not supported) """

    accepted_files = ('.zip', '.xml')

    def convert(self, filepath):
        with zipfile.ZipFile(filepath) as archive:
            data = dict(process_archive(archive))

        return give_me_structure(data)


file_regex = re.compile(
    r'(\d+)__(\d+)__(?P<data>results|qti|tst)_(?P<id>\d+).xml')
task_id_regex = re.compile(r'il_\d+_qst_(?P<task_id>\d+)')

tasks_path = ('./assessment/section')

users = './tst_active/row'
solutions = './tst_solutions/row[@question_fi="%s"]'

lecturer_xpath = ('./MetaData/Lifecycle/Contribute'
                  '[@Role="Author"]/Entity/text()')

types_xpath = ('./item/itemmetadata/qtimetadata/qtimetadatafield/'
               'fieldlabel[text()="QUESTIONTYPE"]/../fieldentry/text()')


def process_qti(tree, only_of_type=('assSourceCode',), **kwargs):
    tasks = tree.xpath(tasks_path)[0]

    titles = tasks.xpath('./item/@title')
    types = tasks.xpath(types_xpath)
    ids = [re.search(task_id_regex, ident).group('task_id')
           for ident in tasks.xpath('./item/@ident')]
    texts = ['\n'.join(flow.xpath('./material/mattext/text()'))
             for flow in tasks.xpath('./item/presentation/flow')]

    return {id: {'title': title, 'text': text, 'type': type}
            for id, type, title, text in zip(ids, types, titles, texts)
            if not only_of_type or type in only_of_type}


def process_users(results_tree):
    return [dict(row.attrib) for row in results_tree.xpath(users)]


def convert_code(text):
    return base64.b64decode(text).decode('utf-8')


def process_solutions(results_tree, task_id):
    return {row.attrib['active_fi']: convert_code(row.attrib['value1'])
            for row in results_tree.xpath(solutions % task_id)}


def process_results(tree, qti=(), **kwargs):
    questions = qti
    users = process_users(tree)
    id2user = {user['active_id']: user for user in users}
    for user in users:
        user['submissions'] = []
    for question_key, question in questions.items():
        solutions = process_solutions(tree, question_key)
        for user_id, solution in solutions.items():
            id2user[user_id]['submissions'].append({'type': question['title'],
                                                    'code': solution,
                                                    'tests': {}})
    return users


def process_tst(tree):
    title = tree.xpath('./MetaData/General/Title/text()')
    lecturer = tree.xpath(lecturer_xpath)
    return {'exam': title[0], 'author': lecturer[0]}


def eval_file(archive, match, cache):
    funcname = 'process_' + match.group('data')
    with archive.open(match.string) as datafile:
        tree = etree.parse(datafile)
        return globals()[funcname](tree, **cache)


def process_archive(archive):
    files = {match.group('data'): match
             for match in (re.search(file_regex, name)
                           for name in archive.NameToInfo)
             if match}

    order = ('tst', 'qti', 'results')
    cache = {}

    for key in order:
        cache[key] = eval_file(archive, files[key], cache)

    return cache


def add_meta(base, data):
    base.update(data['tst'])


def add_tasks(base, data):
    base['tasks'] = list(data['qti'].values())


ignore_user_fields = ("user_fi",
                      "active_id",
                      "usr_id",
                      "anonymous_id",
                      "test_fi",
                      "lastindex",
                      "tries",
                      "submitted",
                      "submittimestamp",
                      "tstamp",
                      "user_criteria",)


def add_users(base, data):
    for userdata in data['results']:
        userdata['identifier'] = userdata['user_fi']
        userdata['username'] = userdata['user_fi']
        for field in ignore_user_fields:
            userdata.pop(field)
    base['students'] = data['results']


def give_me_structure(data):
    base = {}

    add_meta(base, data)
    add_tasks(base, data)
    add_users(base, data)

    return base



import argparse
import base64
import functools
import getpass
import json
import logging
import os
from typing import Any, Callable, Dict, List, Sequence, Union

from lib import Converter
from xkcdpass import xkcd_password as xp

try:
    from cryptography.fernet import Fernet
except ImportError:
    Fernet = None


# ============================== =- Logging -= ============================== #
def setup_logging():
    ''' Make the logger globally available by hide intermediate handler,
    filters and formatter variables '''
    global log

    level = logging.DEBUG if args.verbose else logging.INFO

    log = logging.getLogger(__name__)
    log.setLevel(level)

    # create console handler and formatter
    console = logging.StreamHandler()
    console.setLevel(level)
    formatter = logging.Formatter('[%(levelname)s] %(message)s')

    # add formatter to console handler
    console.setFormatter(formatter)
    log.addHandler(console)


# ============================= =- argparse -= ============================== #
def setup_argparse():
    global args

    def file_exists(parser, filepath: str) -> str:
        if not os.path.isfile(filepath):
            parser.error('Not a file %s' % filepath)
        return filepath

    parser = argparse.ArgumentParser()

    # General purpose arguments
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help='enable verbose logging (Level: DEBUG)')

    # Input output files
    parser.add_argument(
        'input',
        metavar='DATA',
        type=lambda f: file_exists(parser, f),
        help='an Ilias course export in .zip or .xls format')
    parser.add_argument(
        'output',
        metavar='OUTFILE',
        help='destination of converter output (JSON)')

    # Post-processor flags
    remove_personal = parser.add_mutually_exclusive_group()
    remove_personal.add_argument(
        '-e', '--encrypt',
        action='store_true',
        help='''strip all personal information and provide decryption key
                (AES 128-bit, CBC mode, PKCS7 for padding, HMAC with SHA-256
                for integrity)'''
    )
    remove_personal.add_argument(
        '-d', '--decrypt',
        action='store_true',
        help='Reverse previous AES encryption.')
    remove_personal.add_argument(
        '-a', '--anonymous',
        action='store_true',
        help='replace personal information and create a reversing table')
    parser.add_argument(
        '-t', '--personal-secret-table',
        help='where to store personal information (CSV)',
    )
    parser.add_argument(
        '-m', '--add-meta',
        action='store_true',
        help='add meta information (lecturer, course title)'
    )
    parser.add_argument(
        '--verify',
        action='store_true',
        default=True,
        help='asserts that output data will be in a certain format'
    )
    parser.add_argument(
        '-r', '--readable-code',
        action='store_true',
        help='make student code readable by inserting artificial line breaks')

    args = parser.parse_args()

    if (args.decrypt or args.encrypt) and Fernet is None:
        parser.error('To use AES encryption, install cryptography via pip')

    if args.anonymous != (args.personal_secret_table is not None):
        parser.error('Please specify where to write the mapping (see -t)')


# ========================== =- General Purpose -= ========================== #
def compose(*functions: Sequence[Callable]) -> Callable:
    """ Standard function composition. Takes a Sequence of functions [f, g, h, ...]
    and returns the composite function i(x) = f(g(h(x))). There are no checks
    that validate if domain and image of these functions are compatible. """
    return functools.reduce(lambda f,
                            g: lambda x: f(g(x)),
                            functions,
                            lambda x: x)


def abort(message='Bye.'):
    ''' In case anything goes wrong. Basically a dump wrapper around exit '''
    log.info(message)
    exit(1)


# ========================== =- Post processors -= ========================== #
def student_replacer(processor):
    ''' A simple decorator that is used to remove students and put them back in
    when the preprocessor is dome with them'''

    @functools.wraps(processor)
    def processor_closure(structured_data: Dict[str, Any]) -> Dict[str, Any]:
        students = structured_data.pop('students')
        students_replacement = processor(students)
        structured_data['students'] = students_replacement
        return structured_data

    return processor_closure


def do_add_meta(structured_data: Dict[str, Any]) -> Dict[str, Any]:
    ''' Asks the user for metadata about the exam '''
    structured_data['author'] = input('[Q] author: ')
    structured_data['exam'] = input('[Q] course title: ')
    return structured_data


def do_verify(structured_data: Dict[str, Any]) -> Dict[str, Any]:
    ''' The is the testable specification of the format that is output by
    hector. Since multiple formats are compiled into this one verification is
    on by default. The impact on performance is neglectable. '''
    def assert_submission(submission):
        assert 'code' in submission, 'A submission needs code'
        assert type(submission['code']) in [str, list], 'Code is readable'
        assert 'type' in submission, 'A submission has to be of some type'
        assert 'tests' in submission, 'A tests dict has to be present.'

    def assert_student(student):
        log.debug('asserting %s (%d)' % (student['fullname'],
                                         len(student['submissions'])))
        assert 'fullname' in student, 'Student needs a name %s' % student
        assert 'identifier' in student, 'Student needs a unique identifier'
        assert 'username' in student, 'Student needs a unique username'

    def base_assert():
        assert 'students' in structured_data, 'No students found'
        assert 'tasks' in structured_data, 'No tasks found'

    def assert_task(task):
        assert 'type' in task, 'Task has no type'
        assert 'title' in task, 'Task must have a title'

    try:
        base_assert()
        students = structured_data['students']
        tasks = structured_data['tasks']
        number_of_submissions = len(tasks)

        for task in tasks:
            try:
                assert_task(task)
            except AssertionError as err:
                raise err

        for student in students:

            try:
                assert_student(student)
                assert number_of_submissions == len(student['submissions']), \
                    '%s does not have enough submissions' % student['fullname']
                for submission in student['submissions']:

                    try:
                        assert_submission(submission)
                    except AssertionError as err:
                        log.warn(err)

            except AssertionError as err:
                log.warn(err)

    except AssertionError as err:
        log.warn(err)

    return structured_data


@student_replacer
def do_readable_code(students: Dict[str, Union[str, List]]):
    for student in students:
        for submission in student['submissions']:
            submission['code'] = submission['code'].split('\n')
    return students


@student_replacer
def do_anonymous(students: Dict[str, Union[str, List]]):
    ''' Recreates most of the data and includes fields over a whitelist
    therefore ensuring that no personal information remains in the data '''
    DELIMITER = '-'
    wordfile = xp.locate_wordfile()
    words = xp.generate_wordlist(wordfile=wordfile,
                                 min_length=7,
                                 max_length=7)

    def get_random_xkcd_identifier():
        return xp.generate_xkcdpassword(words, numwords=2, delimiter=DELIMITER)

    reverser = {get_random_xkcd_identifier(): s for s in students}
    students_anonymous = [{
        'fullname': ' '.join(w[0].capitalize() + w[1:]
                             for w in pseudo_identifier.split(DELIMITER)),
        'identifier': pseudo_identifier,
        'username': pseudo_identifier,
        'submissions': student['submissions']
    } for pseudo_identifier, student in reverser.items()]

    with open(args.personal_secret_table, 'w') as out:
        print('key;previous identifier;fullname', file=out)
        print('\n'.join('%s;%s;%s' % (anonymous_key,
                                      data['identifier'],
                                      data['fullname'])
                        for anonymous_key, data in reverser.items()), file=out)

    return students_anonymous


@student_replacer
def do_encrypt(students):

    # Init Crypto. See the module documentation on what actually happens here,
    # then read all about those methods and then go study number theory. Never
    # roll your own custom crypto ;-)
    key = Fernet.generate_key()
    aes = Fernet(key)

    def encrypt(clear: str) -> str:
        return base64.b64encode(aes.encrypt(clear.encode())).decode('utf-8')

    output_the_key_to_the_user(key)
    return transform(students, encrypt)


@student_replacer
def do_decrypt(students):

    def decrypt(cipher: str) -> str:
        return aes.decrypt(base64.b64decode(cipher.encode())).decode('utf-8')

    try:
        key = getpass.getpass('[Q] Give me the decryption key: ')
        aes = Fernet(key)

        return transform(students, decrypt)
    except Exception as err:
        abort('Your key is bad (%s).' % err)


# ======================= =- Post processor helper -= ======================= #
def transform(students, function):
    return [
        {'fullname': function(student['fullname']),
         'identifier': function(student['identifier']),
         'username': function(student['username']),
         'submissions': student['submissions']} for student in students
    ]


def output_the_key_to_the_user(key: bytes):

    def to_file(filepath: str):
        with open(filepath, 'wb') as file:
            file.write(key)
        log.info('Key written to %s. Keep it safe.', filepath)

    def to_stdout():
        print('Encrypted and signed. Key this key safe or bad things happen')
        print('   --------->> %s <<---------   ' % key.decode('latin-1'))

    output = input('[Q] The data has been encrypted. ' +
                   'Where should I put the key? (stdout) ') or 'stdout'

    if output == 'stdout':
        to_stdout()

    elif not os.path.exists(output):
        to_file(output)

    elif os.path.isfile(output):
        confirm = input('[Q] File exists. Want to override? (Y/n)') or 'y'
        if confirm.lower().startswith('y'):
            to_file(output)
        else:
            abort('No data was written. Bye.')

    else:
        log.error('I cannot write to %s.', output)
        abort()


def get_active_postprocessors():
    postprocessor_order = (
        do_add_meta,
        do_verify,
        do_readable_code,
        do_anonymous,
        do_encrypt,
        do_decrypt
    )

    return (p for p in postprocessor_order
            if getattr(args, p.__name__.split('do_')[1]))


# ============================== =- Hektor -= =============================== #
def _preprocessing(filepath: str) -> str:
    return filepath


def _processing(filepath: str) -> Dict[str, Any]:
    ''' Find the first apropriate converter and run pass it the path to the
    datafile. '''
    try:
        return next(converter().convert(filepath)
                    for converter in Converter.implementations()
                    if converter.accept(filepath))
    except StopIteration as err:
        log.error('No suitable converter found. Accepting only %s' %
                  ', '.join(f
                            for c in Converter.implementations()
                            for f in c.accepted_files))
        abort('Program stopped prematurely. No data was written. Bye.')


def _postprocessing(structured_data: Dict[str, Any]) -> Dict[str, Any]:
    return compose(*get_active_postprocessors())(structured_data)


def main():
    setup_argparse()
    setup_logging()

    log.debug('Active post processors %s', list(get_active_postprocessors()))

    processing = compose(_postprocessing, _processing, _preprocessing)
    data = processing(args.input)

    destination = args.output.split('.json')[0] + '.json'
    with open(destination, 'w') as output:
        json.dump(data, output, indent=2, sort_keys=True)

    log.info('Wrote exam data to %s', destination)


if __name__ == '__main__':
    main()



#!/usr/local/bin/python3
""" a simple script that converts ilias exam output to readable json

usage: convert.py [-h] [-u USERNAMES] [-n NUMBER_OF_TASKS] INFILE OUTFILE

positional arguments:
  INFILE                Ilias exam data
  OUTFILE               Where to write the final file

optional arguments:
  -h, --help            show this help message and exit
  -u USERNAMES, --usernames USERNAMES
                        a json dict matno -> email
  -n NUMBER_OF_TASKS, --NUMBER_OF_TASKS NUMBER_OF_TASKS
                        Where to write the final file


Author: Jan Maximilian Michal
Date: 30 March 2017
"""

import json
import os
import re
import urllib.parse
from collections import defaultdict, namedtuple

import lib.generic
from xlrd import open_workbook


class XLSConverter(lib.generic.Converter):
    """docstring for XLSConverter"""

    accepted_files = ('.xls',)

    def convert(self, filepath):
        return converter(filepath)


# one user has one submission (code) per task
# yes, I know it is possible to name match groups via (?P<name>) but
# I like this solution better since it gets the job done nicely
user_t = namedtuple('user_head', 'name matrikel_no')

# one task has a title and id and hpfly code
task_head_re = re.compile(r'^Quellcode Frage (?P<title>.*?) ?(\d{8})?$')

# nor parsing the weird mat no
matno_re = re.compile(r'^(?P<matrikel_no>\d+)-(\d+)-(\d+)$')

TABWIDTH = 4


def converter(infile, usernames=None, number_of_tasks=0):

    # Modify these iterators in order to change extraction behaviour

    def sheet_iter_meta(sheet, silent=True):
        """ yield first and second col entry as tuple of (name, matnr) """
        for row in (sheet.row(i) for i in range(1, sheet.nrows)):
            match = re.search(matno_re, row[1].value)
            if match:
                if not silent and len(match.group('matrikel_no')) != 8:
                    print('[WARN] %s has odd matrikelno %s' %
                          (row[0].value, match.group('matrikel_no')))
                yield row[0].value, match.group('matrikel_no')
            else:
                if not silent:
                    print('[WARN] could not parse row %s' % row[0])
                yield row[0].value, row[1].value

    def sheet_iter_data(sheet):
        """ yields all source code title and code tuples """
        def row(i):
            return sheet.row(i)
        for top, low in ((row(i), row(i + 1)) for i in range(sheet.nrows - 1)):
            if any(map(lambda c: c.ctype, top)) and 'Quell' in top[0].value:
                yield (' '.join(c.value for c in top),
                       ' '.join(c.value for c in low))

    # meta sheet contains ilias names usernames etc - data contains code
    meta, *data = open_workbook(infile, open(os.devnull, 'w')).sheets()

    # nice!
    name2mat = dict(sheet_iter_meta(meta, silent=False))
    assert len(name2mat) == len(data), '{} names != {} sheets'.format(len(name2mat), len(data))  # noqa

    # from xls to lists and namedtuples
    # [ [user0, task0_h, code0, ..., taskn, coden ], ..., [...] ]
    root = []
    tasks = {}
    for user, sheet in zip(sheet_iter_meta(meta), data):
        root.append([user_t(*user)])
        for task, code in sheet_iter_data(sheet):
            task = re.search(task_head_re, task)
            task_title = task.group('title')
            tasks[task_title] = {
                'title': task_title,
                'type': 'SourceCode'
            }
            root[-1].append(task.group('title'))
            root[-1].append(urllib.parse
                            .unquote(code)
                            .replace('\t', ' ' * TABWIDTH))

    if number_of_tasks:
        for (user, *task_list) in sorted(root, key=lambda u: u[0].name):
            assert len(task_list) == number_of_tasks * 2

    mat_to_email = defaultdict(str)
    if usernames:
        with open(usernames) as data:
            mat_to_email.update(json.JSONDecoder().decode(data.read()))

    def get_username(user):
        if name2mat[user.name] in mat_to_email:
            return mat_to_email[name2mat[user.name]].split('@')[0]
        return ''.join(filter(str.isupper, user.name)) + name2mat[user.name]

    usernames = {user.name: get_username(user) for (user, *_) in root}

    return {
        'students': [{
            'fullname': user.name,
            'username': usernames[user.name],
            'email': mat_to_email[name2mat[user.name]],
            'identifier': name2mat[user.name],
            'submissions': [{
                "type": task,
                "code": code,
                "tests": {},
            } for task, code in zip(task_list[::2], task_list[1::2])]
        } for (user, *task_list) in sorted(root, key=lambda u: u[0].name)],
        'tasks': list(tasks.values())
    }


def write_to_file(json_dict, outfile):
    # just encode python style
    with open(outfile, "w") as out:
        json.dump(json_dict, out, indent=2)

    print("Wrote data to %s. Done." % outfile)



import json

import lib.generic


class JSONIdentityConverter(lib.generic.Converter):
    """ This serves as an identity if you wish to import a json file
    that you generated earlier with hektor and you now want to run a
    preprocessor on it. """

    accepted_files = ('.json',)

    def convert(self, filepath):
        with open(filepath) as json_input:
            return json.load(json_input)



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='hektor',
    version='0.3.6',
    description='A QTI-XML/XLS to JSON converter for humans',
    author='Jan Maximilian Michal',
    author_email='mail@janmax.org',
    url='https://gitlab.gwdg.de/j.michal/hektor',
    license='MIT',
    scripts=['bin/hektor'],
    install_requires=["lxml~=4.1.1",
                      "xlrd~=1.1.0",
                      "xkcdpass~=1.16.0"],
    py_modules=['hektor'],
    packages=['lib']
)



import abc


def all_subclasses(cls):
    return cls.__subclasses__() \
        + [g for s in cls.__subclasses__() for g in all_subclasses(s)]


class Converter(metaclass=abc.ABCMeta):
    """ A base class if we incorporate more converters in the future. New
    implementations need to be registered in this modules __init__.py """

    @abc.abstractmethod
    def convert(self):
        pass

    @property
    @abc.abstractclassmethod
    def accepted_files(cls):
        pass

    @classmethod
    def implementations(cls):
        return all_subclasses(cls)

    @classmethod
    def accept(cls, filepath):
        return any(filepath.endswith(ending) for ending in cls.accepted_files)



# New modules need to be registered here

from lib.generic import Converter  # noqa
from lib.qti import QTIConverter  # noqa
from lib.xls import XLSConverter  # noqa
from lib.identity import JSONIdentityConverter  # noqa
